==> PyArweave-0.0.1.tar.gz: PyArweave-0.0.1/ar/arweave_lib.py <==
# This file is part of PyArweave.
# 
# PyArweave is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 2 of the License, or (at your option) any later
# version.
# 
# PyArweave is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# PyArweave. If not, see <https://www.gnu.org/licenses/>.

import json
import os
import io
import requests
import logging
import hashlib
import psutil
import arrow
import nacl.bindings
from jose import jwk
from jose.utils import base64url_encode, base64url_decode, base64
from jose.backends.cryptography_backend import CryptographyRSAKey
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_PSS
from Crypto.Hash import SHA256
from .utils import (
    winston_to_ar,
    ar_to_winston,
    owner_to_address,
    create_tag,
    encode_tag,
    decode_tag,
    base64url_decode
)
from .deep_hash import deep_hash
from .merkle import compute_root_hash, generate_transaction_chunks

logger = logging.getLogger(__name__)

TRANSACTION_DATA_LIMIT_IN_BYTES = 2000000
API_URL = "https://arweave.net"


class ArweaveTransactionException(Exception):
    pass

class Peer(object):
    # peer api is at https://docs.arweave.org/developers/server/http-api

    def __init__(self, api_url = API_URL):
        self.api_url = api_url

    def _get(self, format, *params):
        url = format.format(self.api_url, *params)

        response = requests.get(url)

        if response.status_code == 200:
            return response
        else:
            logger.error(response.text)
            raise ArweaveTransactionException(response.text)

    def _post(self, json_data, format, *params):
        url = format.format(self.api_url, *params)

        headers = {'Content-Type': 'application/json', 'Accept': 'text/plain'}

        response = requests.post(url, data=json_data, headers=headers)

        logger.debug("{}\n\n{}".format(response.text, self.json_data))

        if response.status_code == 200:
            logger.debug("RESPONSE 200: {}".format(response.text))
            return response
        else:
            logger.error("{}\n\n{}".format(response.text, json_data))
            raise ArweaveTransactionException(response.text, json_data)
        

    def tx(self, txid):
        """Get Transaction by ID"""
        tx_response = self._get("{}/tx/{}", txid)
        return tx_response.text

    def tx_status(self, txid):
        """Get Transaction Status"""
        response = self._get("{}/tx/{}/status", txid)
        return json.loads(response.text)

    def tx_field(self, txid, field):
        """Get Transaction Field"""
        response = self._get("{}/tx/{}/{}", txid, field)
        return json.loads(response.text)

    def data(self, txid):
        """
        Get the decoded data from a transaction.

        The content type is not preserved.

        The transaction is pending: Pending
        The provided transaction ID is not valid or the field name is not valid: Invalid hash.
        A transaction with the given ID could not be found: Not Found.
        """
        response = self._get("{}/{}/", txid)
        return response.content

    def price(self, data_size=0, target_address=None):
        """Get Transaction Price"""
        if target_address:
            response = self._get("{}/price/{}/{}", data_size, target_address)
        else:
            response = self._get("{}/price/{}", data_size)
        return response.text

    def send(self, json_data):
        """Submit a Transaction"""
        response = self._post(json_data)
        return response.text

    def wallet_balance(self, address):
        """Get a Wallet Balance"""
        response = self._get("{}/wallet/{}/balance", address)
        return response.text

    def tx_anchor(self):
        # TODO: the docs say this should be https://arweave.net/wallet/{address}/last_tx
        """Get Last Transaction ID"""
        last_tx_response = self._get("{}/tx_anchor")
        return last_tx_response.text

    def block_hash(self, block_hash):
        """Get Block by ID"""
        response = self._get("{}/block/hash/{}", block_hash)
        return json.loads(response.text)

    def info(self):
        """Network Info"""
        response = self._get("{}/info")
        return json.loads(response.text)

    def peers(self):
        """Peer list"""
        response = self._get("{}/peers")
        return json.loads(response.text)

    def chunk(self, json_data):
        """
        Upload Data Chunks

        {
          "data_root": "<Base64URL encoded data merkle root>",
          "data_size": "a number, the size of transaction in bytes",
          "data_path": "<Base64URL encoded inclusion proof>",
          "chunk": "<Base64URL encoded data chunk>",
          "offset": "<a number from [start_offset, start_offset + chunk size), relative to other chunks>"
        }
        """
        response = self._post("{}/chunk")

    def tx_data(self, txid):
        """Get Transaction Data"""
        response = self._get("{}/tx/{}/data", txid)
        return response.content

    def tx_offset(self, txid):
        """Get Transaction Offset and Size"""
        response = self._get("{}/tx/{}/offset", txid)
        return response.text


class Wallet(object):
    HASH = 'sha256'

    def _set_jwk_params(self):
        self.jwk_data['p2s'] = ''
        self.jwk = jwk.construct(self.jwk_data, algorithm=jwk.ALGORITHMS.RS256)
        self.rsa = RSA.importKey(self.jwk.to_pem())

        self.owner = self.jwk_data.get('n')
        self.address = owner_to_address(self.owner)

        self.peer = Peer(API_URL)
    @property
    def api_url(self):
        return self.peer.api_url
    @api_url.setter
    def set_api_url(self, api_url):
        self.peer.api_url = api_url

    def __init__(self, jwk_file='jwk_file.json'):
        with open(jwk_file, 'r') as j_file:
            self.jwk_data = json.loads(j_file.read())
        self._set_jwk_params()

    @classmethod
    def from_data(cls, jwk_data):
        wallet = cls.__new__(cls)
        wallet.jwk_data = jwk_data
        wallet._set_jwk_params()
        return wallet

    @property
    def balance(self):
        balance = self.peer.wallet_balance(self.address)
        return winston_to_ar(balance)

    def sign(self, message):
        h = SHA256.new(message)
        signed_data = PKCS1_PSS.new(self.rsa).sign(h)
        return signed_data

    def verify(self):
        pass

    def get_last_transaction_id(self):
        self.last_tx = self.peer.tx_anchor()
        return self.last_tx

class Transaction(object):
    def __init__(self, wallet, **kwargs):
        self.jwk_data = wallet.jwk_data
        self.jwk = jwk.construct(self.jwk_data, algorithm="RS256")
        self.wallet = wallet

        self.id = kwargs.get('id', '')
        self.last_tx = wallet.get_last_transaction_id()
        self.owner = self.jwk_data.get('n')
        self.tags = []
        self.format = kwargs.get('format', 2)

        self.peer = Peer(API_URL)
        self.chunks = None

        data = kwargs.get('data', '')
        self.data_size = len(data)

        if type(data) is bytes:
            self.data = base64url_encode(data)
        else:
            self.data = base64url_encode(data.encode('utf-8'))

        if self.data is None:
            self.data = ''

        self.file_handler = kwargs.get('file_handler', None)
        if self.file_handler:
            self.uses_uploader = True
            self.data_size = os.stat(kwargs['file_path']).st_size
        else:
            self.uses_uploader = False

        if kwargs.get('transaction'):
            self.from_serialized_transaction(kwargs.get('transaction'))
        else:
            self.data_root = ""

            self.data_tree = []

            self.target = kwargs.get('target', '')
            self.to = kwargs.get('to', '')

            if self.target == '' and self.to != '':
                self.target = self.to

            self.quantity = kwargs.get('quantity', '0')
            if float(self.quantity) > 0:
                if self.target == '':
                    raise ArweaveTransactionException(
                        "Unable to send {} AR without specifying a target address".format(self.quantity))

                # convert to winston
                self.quantity = ar_to_winston(float(self.quantity))

            reward = kwargs.get('reward', None)
            if reward is not None:
                self.reward = reward

            self.signature = ''
            self.status = None
    @property
    def api_url(self):
        return self.peer.api_url
    @api_url.setter
    def set_api_url(self, api_url):
        self.peer.api_url = api_url

    def from_serialized_transaction(self, transaction_json):
        if type(transaction_json) == str:
            self.load_json(transaction_json)
        else:
            raise ArweaveTransactionException(
                "Please supply a string containing json to initialize a serialized transaction")

    def get_reward(self, data_size, target_address=None):
        reward = self.peer.price(data_size, target_address)
        return winston_to_ar(reward)

    def add_tag(self, name, value):
        tag = create_tag(name, value, self.format == 2)
        self.tags.append(tag)

    def encode_tags(self):
        tags = []
        for tag in self.tags:
            tags.append(encode_tag(tag))

        self.tags = tags

    def sign(self):
        data_to_sign = self.get_signature_data()

        raw_signature = self.wallet.sign(data_to_sign)

        self.signature = base64url_encode(raw_signature)

        self.id = base64url_encode(hashlib.sha256(raw_signature).digest())

        if type(self.id) == bytes:
            self.id = self.id.decode()

    def get_signature_data(self):
        self.reward = self.get_reward(self.data_size, target_address=self.target if len(self.target) > 0 else None)

        if int(self.data_size) > 0 and self.data_root == "" and not self.uses_uploader:
            if type(self.data) == str:
                root_hash = compute_root_hash(io.BytesIO(base64url_decode(self.data.encode('utf-8'))))

            if type(self.data) == bytes:
                root_hash = compute_root_hash(io.BytesIO(base64url_decode(self.data)))

            self.data_root = base64url_encode(root_hash)

        if self.format == 1:
            tag_str = ""

            for tag in self.tags:
                name, value = decode_tag(tag)
                tag_str += "{}{}".format(name.decode(), value.decode())

            owner = base64url_decode(self.jwk_data['n'].encode())
            target = base64url_decode(self.target)
            data = base64url_decode(self.data)
            quantity = self.quantity.encode()
            reward = self.reward.encode()
            last_tx = base64url_decode(self.last_tx.encode())

            signature_data = owner + target + data + quantity + reward + last_tx + tag_str.encode()

        if self.format == 2:
            if self.uses_uploader:
                self.prepare_chunks()

            tag_list = [[tag['name'].encode(), tag['value'].encode()] for tag in self.tags]

            signature_data_list = [
                "2".encode(),
                base64url_decode(self.jwk_data['n'].encode()),
                base64url_decode(self.target.encode()),
                str(self.quantity).encode(),
                self.reward.encode(),
                base64url_decode(self.last_tx.encode()),
                tag_list,
                str(self.data_size).encode(),
                base64url_decode(self.data_root)]

            signature_data = deep_hash(signature_data_list)

        return signature_data

    def send(self):
        try:
            self.peer.send(self.json_data)
        except ArweaveTransactionException:
            pass

        return self.last_tx

    def to_dict(self):

        if self.data is None:
            self.data = ''

        data = {
            'data': self.data.decode() if type(self.data) == bytes else self.data,
            'id': self.id.decode() if type(self.id) == bytes else self.id,
            'last_tx': self.last_tx,
            'owner': self.owner,
            'quantity': self.quantity,
            'reward': self.reward,
            'signature': self.signature.decode(),
            'tags': self.tags,
            'target': self.target
        }

        if self.format == 2:
            self.encode_tags()
            data['tags'] = self.tags
            data['format'] = 2
            if len(self.data_root) > 0:
                data['data_root'] = self.data_root.decode()
            else:
                data['data_root'] = ""
            data['data_size'] = str(self.data_size)
            data['data_tree'] = []

        return data

    @property
    def json_data(self):
        data = self.to_dict()

        json_str = json.dumps(data)

        logger.debug(json_str)

        return json_str.replace(' ', '')

    def get_status(self):
        try:
            self.status = self.peer.tx_status(self.id)
        except ArweaveTransactionException:
            self.status = "PENDING"
        return self.status

    def get_transaction(self):
        try:
            self.load_json(self.peer.tx(self.id))
        except ArweaveTransactionException as exception:
            pass

    def get_price(self):
        try:
            price = self.peer.price(self.data_size)
            return winston_to_ar(price)
        except TransactionException as exception:
            pass

    def get_data(self):
        self.data = self.peer.data(self.id)
        return self.data

    def load_json(self, json_str):
        json_data = json.loads(json_str)

        self.data = json_data.get('data', '')
        self.last_tx = json_data.get('last_tx', '')
        self.owner = json_data.get('owner', '')
        self.quantity = json_data.get('quantity', '')
        self.reward = json_data.get('reward', '')
        self.signature = json_data.get('signature', '')
        self.tags = [decode_tag(tag) for tag in json_data.get('tags', [])]
        self.target = json_data.get('target', '')
        self.data_size = json_data.get('data_size', '0')
        self.data_root = json_data.get('data_root', '')
        self.data_tree = json_data.get('data_tree', [])

        logger.debug(json_data)

    def prepare_chunks(self):
        if not self.chunks:
            self.chunks = generate_transaction_chunks(self.file_handler)
            self.data_root = base64url_encode(self.chunks.get('data_root'))

        if not self.chunks:
            self.chunks = {
                "chunks": [],
                "data_root": b'',
                "proof": []
            }

            self.data_root = ''

    def get_chunk(self, idx):
        if self.chunks is None:
            raise ArweaveTransactionException("Chunks have not been prepared")

        proof = self.chunks.get('proofs')[idx]
        chunk = self.chunks.get('chunks')[idx]

        self.file_handler.seek(chunk.min_byte_range)

        chunk_data = self.file_handler.read(chunk.data_size)

        return {
            "data_root": self.data_root.decode(),
            "data_size": str(self.data_size),
            "data_path": base64url_encode(proof.proof),
            "offset": str(proof.offset),
            "chunk": base64url_encode(chunk_data)
        }


def arql(wallet, query):
    """
    Creat your query like so:
    query = {
        "op": "and",
          "expr1": {
            "op": "equals",
            "expr1": "from",
            "expr2": "hnRI7JoN2vpv__w90o4MC_ybE9fse6SUemwQeY8hFxM"
          },
          "expr2": {
            "op": "or",
            "expr1": {
              "op": "equals",
              "expr1": "type",
              "expr2": "post"
            },
            "expr2": {
              "op": "equals",
              "expr1": "type",
              "expr2": "comment"
            }
          }
    :param wallet:
    :param query:
    :return list of Transaction instances:
    """

    data = json.dumps(query)
    headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}
    response = requests.post("{}/arql".format(API_URL), data=data, headers=headers)

    if response.status_code == 200:
        transaction_ids = json.loads(response.text)

        return transaction_ids

    return None


def arql_with_transaction_data(wallet, query):
    """
    Creat your query like so:
    query = {
        "op": "and",
          "expr1": {
            "op": "equals",
            "expr1": "from",
            "expr2": "hnRI7JoN2vpv__w90o4MC_ybE9fse6SUemwQeY8hFxM"
          },
          "expr2": {
            "op": "or",
            "expr1": {
              "op": "equals",
              "expr1": "type",
              "expr2": "post"
            },
            "expr2": {
              "op": "equals",
              "expr1": "type",
              "expr2": "comment"
            }
          }
    :param wallet:
    :param query:
    :return list of Transaction instances:
    """

    transaction_ids = arql(wallet, query)
    if transaction_ids:
        transactions = []
        for transaction_id in transaction_ids:
            tx = Transaction(wallet, id=transaction_id)
            tx.get_transaction()
            tx.get_data()

            transactions.append(tx)

    return None

==> PyArweave-0.1.3.tar.gz: PyArweave-0.1.3/ar/arweave_lib.py <==
# This file is part of PyArweave.
# 
# PyArweave is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 2 of the License, or (at your option) any later
# version.
# 
# PyArweave is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with
# PyArweave. If not, see <https://www.gnu.org/licenses/>.

import json
from .peer import Peer

from . import DEFAULT_API_URL

TRANSACTION_DATA_LIMIT_IN_BYTES = 2000000

def arql(wallet, query):
    """
    Creat your query like so:
    query = {
        "op": "and",
          "expr1": {
            "op": "equals",
            "expr1": "from",
            "expr2": "hnRI7JoN2vpv__w90o4MC_ybE9fse6SUemwQeY8hFxM"
          },
          "expr2": {
            "op": "or",
            "expr1": {
              "op": "equals",
              "expr1": "type",
              "expr2": "post"
            },
            "expr2": {
              "op": "equals",
              "expr1": "type",
              "expr2": "comment"
            }
          }
    :param wallet:
    :param query:
    :return list of Transaction instances:
    """

    return Peer(DEFAULT_API_URL).arql(query)


def arql_with_transaction_data(wallet, query):
    """
    Creat your query like so:
    query = {
        "op": "and",
          "expr1": {
            "op": "equals",
            "expr1": "from",
            "expr2": "hnRI7JoN2vpv__w90o4MC_ybE9fse6SUemwQeY8hFxM"
          },
          "expr2": {
            "op": "or",
            "expr1": {
              "op": "equals",
              "expr1": "type",
              "expr2": "post"
            },
            "expr2": {
              "op": "equals",
              "expr1": "type",
              "expr2": "comment"
            }
          }
    :param wallet:
    :param query:
    :return list of Transaction instances:
    """

    transaction_ids = arql(wallet, query)
    if transaction_ids:
        transactions = []
        for transaction_id in transaction_ids:
            tx = Transaction(wallet, id=transaction_id)
            tx.get_transaction()
            tx.get_data()

            transactions.append(tx)

    return None
